////////////////////////////////////////////////////////////////////////////////
//
// OpenGL support - Unit:
//
// Include support for OpenGL preview of NormalTexture and AnimatedTextures
//
//
// (c) 2001, by Jesco Topp
////////////////////////////////////////////////////////////////////////////////
unit OpenGLPreview;

interface

uses Windows, Dialogs, classes, graphics, jpeg, OpenGL, GLAux, Objects, PackageFormat;

//------------------------------------------------------------------------------
// Import some functions from OpenGL32.dll

procedure glGenTextures(n: GLsizei; var textures: GLuint); stdcall; external opengl32;
procedure glBindTexture(target: GLenum; texture: GLuint); stdcall; external opengl32;

//------------------------------------------------------------------------------
// Header

function  CreateOpenGL( const WindowHandle : HWND; Width, Height : integer ) : Boolean;
function  KillOpenGL : Boolean;
function  LoadTextures( GroupIndex, SelTexture : integer; IsAnimated : Boolean ) : TStrings;
// Draw functions
procedure DrawGL;
procedure ZoomGL( Factor : real );

//------------------------------------------------------------------------------
// global OpenGL variables

var
  h_WND         : HWND;
  h_DC          : HDC;
  h_RC          : HGLRC;
  Texture       : array [0..99] of GLuint;    // Storage for Textures (if Animated then for each frame)
  z             : GLfloat;
  TexName       : string;
  TexWidth      : integer;
  TexHeight     : integer;
  Animate       : Boolean;
  CurFrame      : integer;


//------------------------------------------------------------------------------

implementation

//------------------------------------------------------------------------------
// Implementation


//------------------------------------------------------------------------------
// Init and destroy

// Create a new OpenGL rendering context (needs a HWND, width and height)
function CreateOpenGL;
var
  Pixelformat     : GLuint;                    // Holds The Result After Searching For A Match
  pfd             : TPixelFormatDescriptor;    // Tells Windows How We Want Things To Be
  fwidth, fheight : GLfloat;
begin
  // copy temporary handle to permanent one in this unit
  h_WND := WindowHandle;

  with pfd do    //Tells Windows How We Want Things To Be
  begin
    nSize:= SizeOf( TPIXELFORMATDESCRIPTOR ); // Size Of This Pixel Format Descriptor
    nVersion:= 1;                            // Version Number (?)
    dwFlags:= PFD_DRAW_TO_WINDOW             // Format Must Support Window
      or PFD_SUPPORT_OPENGL                  // Format Must Support OpenGL
      or PFD_DOUBLEBUFFER;                   // Must Support Double Buffering
    iPixelType:= PFD_TYPE_RGBA;              // Request An RGBA Format
    cColorBits:= 16;                         // Select Our Color Depth
    cRedBits:= 0;                            // Color Bits Ignored
    cRedShift:= 0;
    cGreenBits:= 0;
    cBlueBits:= 0;
    cBlueShift:= 0;
    cAlphaBits:= 0;                          // No Alpha Buffer
    cAlphaShift:= 0;                         // Shift Bit Ignored
    cAccumBits:= 0;                          // No Accumulation Buffer
    cAccumRedBits:= 0;                       // Accumulation Bits Ignored
    cAccumGreenBits:= 0;
    cAccumBlueBits:= 0;
    cAccumAlphaBits:= 0;
    cDepthBits:= 16;                         // 16Bit Z-Buffer (Depth Buffer)
    cStencilBits:= 0;                        // No Stencil Buffer
    cAuxBuffers:= 0;                         // No Auxiliary Buffer
    iLayerType:= PFD_MAIN_PLANE;             // Main Drawing Layer
    bReserved:= 0;                           // Reserved
    dwLayerMask:= 0;                         // Layer Masks Ignored
    dwVisibleMask:= 0;
    dwDamageMask:= 0;
  end;

  h_DC := GetDC(h_WND);
  if ( h_DC = 0 ) then
    ShowMessage('Failed to obtain h_DC handle');
  PixelFormat := ChoosePixelFormat(h_Dc, @pfd);               // Finds The Closest Match To The Pixel Format We Set Above
  if ( PixelFormat = 0 ) then
    ShowMessage('Failed to choose PixelFormat');
  if (not SetPixelFormat(h_Dc,PixelFormat,@pfd)) then         // Are We Able To Set The Pixelformat?
    ShowMessage('Could not set PixelFormat');
  h_Rc := wglCreateContext(h_Dc);                             // Are We Able To Get A Rendering Context?
  if (h_Rc=0) then
    ShowMessage('Could not create rendering context');
  if (not wglMakeCurrent(h_Dc, h_Rc)) then                    // Are We Able To Activate The Rendering Context?
    ShowMessage('Could not make rendering context current');

  // ReSizeGL
  fwidth := Width;
  fheight := Height;

  glViewport(0, 0, Width, Height);   // Reset The Current Viewport And Perspective Transformation
  glMatrixMode(GL_PROJECTION);       // Select The Projection Matrix
  glLoadIdentity();                  // Reset The Projection Matrix
  gluPerspective(45.0,fWidth/fHeight,0.1,100.0);// Calculate The Aspect Ratio Of The Window
  glMatrixMode(GL_MODELVIEW);        // Select The Modelview Matrix
  glLoadIdentity;                     //Reset The Modelview Matrix

  //InitGL
  glShadeModel(GL_SMOOTH);			   // Enables Smooth Color Shading
  glClearColor(0.0, 0.0, 0.0, 0.5);		   // Black Background
  glClearDepth(1.0);				   // Depth Buffer Setup
  glEnable(GL_DEPTH_TEST);			   // Enables Depth Testing
  glDepthFunc(GL_LESS);				   // The Type Of Depth Test To Do
  glHint(GL_PERSPECTIVE_CORRECTION_HINT,GL_NICEST);//Realy Nice perspective calculations

  glEnable(GL_TEXTURE_2D);			   // Enable Texture Mapping

  // set initial zoom
  z := -4;

  Result := TRUE;
end;

// Destroy an OpenGL rendering context
function KillOpenGL;
begin
  if h_rc<>0 then                                                               //Is There A Rendering Context?
    begin
      if (not wglMakeCurrent(h_Dc,0)) then                                      //Are We Able To Release Dc and Rc contexts?
        MessageBox(0,'Release of DC and RC failed.',' Shutdown Error',MB_OK or MB_ICONERROR);
      if (not wglDeleteContext(h_Rc)) then                                      //Are We Able To Delete The Rc?
        begin
          MessageBox(0,'Release of Rendering Context failed.',' Shutdown Error',MB_OK or MB_ICONERROR);
          h_Rc:=0;                                                              //Set Rc To Null
        end;
    end;
  if (h_Dc=1) and (releaseDC(h_Wnd,h_Dc)<>0) then                               //Are We Able To Release The Dc?
    begin
      MessageBox(0,'Release of Device Context failed.',' Shutdown Error',MB_OK or MB_ICONERROR);
      h_Dc:=0;                                                                  //Set Dc To Null
    end;
  Result := TRUE;
end;

//------------------------------------------------------------------------------
// Draw

// Draw the OpenGL output
procedure DrawGL;
begin
  glClear(GL_COLOR_BUFFER_BIT or GL_DEPTH_BUFFER_BIT); // Clear The Screen And The Depth Buffer
  glLoadIdentity();		      // Reset The View
  glTranslatef(0.0,0.0,z);

  glBindTexture(GL_TEXTURE_2D, Texture[CurFrame]);

  glBegin(GL_QUADS);
    glNormal3f(  0.0,  0.0,   1.0);   // Normal vector

      glTexCoord2f( 0.0,  1.0);
    glVertex3f( -TexWidth/200,  TexHeight/200,   0.0);   // upper left

      glTexCoord2f( 1.0,  1.0);
    glVertex3f(  TexWidth/200,  TexHeight/200,   0.0);   // upper right

      glTexCoord2f( 1.0,  0.0);
    glVertex3f(  TexWidth/200,  -TexHeight/200,   0.0);   // lower right

      glTexCoord2f( 0.0,  0.0);
    glVertex3f( -TexWidth/200,  -TexHeight/200,   0.0);   // lower left

  glEnd();

  if ( CurFrame + 1 > FrameCount-1 ) then CurFrame := 0 else CurFrame := CurFrame + 1;

  SwapBuffers(h_DC);
end;

// Load current textures and convert them to OpenGL
function LoadTextures;
var
  texture1   : PTAUX_RGBImageRec;
  ConvJPEG   : TJPEGImage;
  ConvBitmap : TBitmap;
  FileName   : string;
  AnimNext   : string;
  i          : integer;
  Properties : TStrings;
begin
  // Init
  Properties := TStringList.Create;
  FrameCount := 0;
  MinFrameRate := 0;
  MaxFrameRate := 0;

  // Load texture (maximum 256*256, because of 3DFX hardware) and save it to disk temporary   (*ugly*)
  UTPackage.LoadNormalTexture(UTPackage.Group[GroupIndex].Entries.Strings[SelTexture], 256);

  ConvJPEG := TJPEGImage.Create;
  ConvJPEG.PixelFormat := jf24bit;
  ConvJPEG.Assign(UTPackage.NormalTexture.B);

  ConvBitmap := TBitmap.Create;
  ConvBitmap.Assign(ConvJPEG);
  ConvBitmap.SaveToFile(Directory + 'OGL.BMP');
  ConvBitmap.Free;
  ConvJPEG.Free;

  // Get texture infos
  TexWidth  := UTPackage.NormalTexture.B.Width;
  TexHeight := UTPackage.NormalTexture.B.Height;
  TexName   := UTPackage.NormalTexture.TextureName;
  UTPackage.NormalTexture.ReadGenericProperties( UTPackage.NameTable, UTPackage.ImportTable, UTPackage.ExportTable, Properties, UTPackage.F );


  // Load Texture
  texture1 := auxDIBImageLoadA(PChar(Directory + 'OGL.BMP'));
  if (not Assigned(texture1)) then
     Halt(1);

  // Create Texture
  glGenTextures(1, texture[0]);
  glBindTexture(GL_TEXTURE_2D, texture[0]);
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
  glTexImage2D(GL_TEXTURE_2D, 0, 3, texture1^.sizeX, texture1^.sizeY, 0, GL_RGB, GL_UNSIGNED_BYTE, texture1^.data);

  // Delete all temporary files
  FileName := Directory + 'OGL.BMP';
  DeleteFile(PChar(FileName));

  // Load following frames if texture is animated
  if ( UTPackage.NormalTexture.IsAnimated = TRUE ) then
  begin
    Animate := TRUE;
    FrameCount := 1;

    // Find AnimNext and load it
    repeat
      AnimNext := UTPackage.NormalTexture.AnimNext;

      if ( UTPackage.NormalTexture.HasMaxFrameRate = TRUE ) then
        MaxFrameRate := UTPackage.NormalTexture.MaxFrameRate;
      if ( UTPackage.NormalTexture.HasMinFrameRate = TRUE ) then
        MinFrameRate := UTPackage.NormalTexture.MinFrameRate;

      for i:=0 to UTPackage.Group[GroupIndex].Entries.Count-1 do
      begin
        if ( UTPackage.Group[GroupIndex].Entries.Strings[i] = AnimNext ) then
        begin
          // Load texture (maximum 256*256, because of 3DFX hardware) and save it to disk temporary   (*ugly*)
          UTPackage.LoadNormalTexture(UTPackage.Group[GroupIndex].Entries.Strings[i], 256);

          ConvJPEG := TJPEGImage.Create;
          ConvJPEG.PixelFormat := jf24bit;
          ConvJPEG.Assign(UTPackage.NormalTexture.B);

          ConvBitmap := TBitmap.Create;
          ConvBitmap.Assign(ConvJPEG);
          ConvBitmap.SaveToFile(Directory + 'OGL.BMP');
          ConvBitmap.Free;
          ConvJPEG.Free;

          // Generic Properties
          UTPackage.NormalTexture.ReadGenericProperties( UTPackage.NameTable, UTPackage.ImportTable, UTPackage.ExportTable, Properties, UTPackage.F );

          // Load Texture
          texture1 := auxDIBImageLoadA(PChar(Directory + 'OGL.BMP'));
          if (not Assigned(texture1)) then
            Halt(1);

          // Create Texture
          glGenTextures(1{FrameCount+1}, texture[FrameCount]);
          glBindTexture(GL_TEXTURE_2D, texture[FrameCount]);
          glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
          glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
          glTexImage2D(GL_TEXTURE_2D, 0, 3, texture1^.sizeX, texture1^.sizeY, 0, GL_RGB, GL_UNSIGNED_BYTE, texture1^.data);

          // Delete all temporary files
          FileName := Directory + 'OGL.BMP';
          DeleteFile(PChar(FileName));

          FrameCount := FrameCount + 1;

          break;
        end;
      end;

      // Additional break condition
      if ( AnimNext = '' ) then
        break;
    until ( AnimNext = TexName );
  end
  else
    Animate := FALSE;


  // Return value
  Result := TStringList.Create;
  Result.AddStrings(Properties);
  //Result.Free;
  Properties.Free;
end;

// Zoom In & Out
procedure ZoomGL;
begin
  // positive values = zoom in
  // negative values = zoom out

  z := z + Factor;
end;

//------------------------------------------------------------------------------

end.
