////////////////////////////////////////////////////////////////////////////////
//
// PkgAux-Unit:
//
// Includes: - procdures for reading INDEX and NAME
//           - TNameTable, TImportTable and TExportTable declarations
//           - TUTColor and TPalette declaration
//           - TAddGroup for .UTX-creation
//
// (c) 2000-2001, by Jesco Topp
////////////////////////////////////////////////////////////////////////////////
unit PkgAux;

interface
uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls;

////////////////////////////////////////////////////////////////////////////////
//functions to read INDEX and NAME

  //INDEX
  function ReadINDEX(var F: file): DWORD; //Read
  procedure WriteINDEX(const Index : DWORD; var F : File); //Write
  function SizeOfINDEX(const Index : DWORD): integer; //Get Size of INDEX
  //NAME
  function ReadNAME(var F: file; PkgVer : integer): string; //Read
  procedure WriteNAME(const Name : string; var F : File); //Write (PkgVer = 69)


////////////////////////////////////////////////////////////////////////////////
//Headertable-objects and ObjectReference-function

type
  TNameTable = Class(TObject)
    ObjectName     : TStrings;
    ObjectFlags    : TStrings;
    Count          : integer;
    Offset         : integer;
    constructor Create;
    procedure Free;
    destructor Destroy; override;
end;

type
  TImportTable = Class(TObject)
    ClassPackage   : TStrings;
    ClassName      : TStrings;
    ImportPackage  : TStrings;
    ObjectName     : TStrings;
    Count          : integer;
    Offset         : integer;
    constructor Create;
    procedure Free;
    destructor Destroy; override;
end;

type
  TExportTable = Class(TObject)
    ExportClass  : TStrings;
    Super        : TStrings;
    ExportPackage: TStrings;
    ObjectName   : TStrings;
    ObjectFlags  : TStrings;
    SerialSize   : TStrings;
    SerialOffset : TStrings;
    Count        : integer;
    Offset       : integer;
    constructor Create;
    procedure Free;
    destructor Destroy; override;
end;

  function  GetObjectReference(const NameTable : TNameTable; ImportTable : TImportTable; ExportTable : TExportTable; Value : DWORD): string;

////////////////////////////////////////////////////////////////////////////////
//Palette and color object

type
  TUTColor = Class(TObject)
    r       : BYTE;
    g       : BYTE;
    b       : BYTE;
    a       : BYTE;
    constructor Create;
    procedure Free;
    destructor Destroy; override;
end;

type
  TPalette = Class(TObject)
    PaletteName    : string;                     //Name of the palette
    PaletteSize    : DWORD;                      //Number of entries
    PaletteOffset  : DWORD;                      //Offset
    Color          : array [0..255] of TUTColor; //unconverted palette-entries
    WinPalette     : HPalette;                   //converted HPALETTE
    constructor Create;
    procedure Free;
    destructor Destroy; override;
  public
    //Loads the under Name specified palette and converts it to a HPALETTE
    procedure Load(const NameTable : TNameTable; ImportTable : TImportTable; ExportTable : TExportTable; var F : File; Name : string; TexturePackage : string); //Read and convert the specified palette
end;

////////////////////////////////////////////////////////////////////////////////
//TAddGroup

type
  TAddGroup = Class(TObject)
    GroupName   : string;
    GroupIndex  : DWORD; //Index of group in ExportTable
    TextureName : TStrings;
    FileName    : TStrings;
    Offset      : TStrings;
    HasMipMaps  : TStrings;
    bMasked     : TStrings;
    constructor Create;
    procedure Free;
    destructor Destroy; override;
  end;

////////////////////////////////////////////////////////////////////////////////

//Read/Write-Exception
type
  TIllegalPackageOperationException = Class(Exception)
  end;

////////////////////////////////////////////////////////////////////////////////

implementation

////////////////////////////////////////////////////////////////////////////////
//functions to read INDEX and NAME

///////////////////////
//Read INDEX-Format
function ReadINDEX;
var
  B0,B1,B2,B3,B4 : BYTE;
  D0,D1,D2,D3,D4 : DWORD;
  INDEX          : DWORD;
  IsNegative     : Boolean;
begin
  B0 := 0;
  B1 := 0;
  B2 := 0;
  B3 := 0;
  B4 := 0;

  BlockRead(F, B0, 1);
  if (B0 > $80) then
  begin
    B0 := B0 - $80;
    IsNegative := TRUE;
  end
  else
    IsNegative := FALSE;

  //Read Following Bytes
  if (B0 >= $40) then
  begin
    B0 := B0 - $40;
    BlockRead(F, B1, 1);
    if (B1 >= $80) then
    begin
      B1 := B1 - $80;
      BlockRead(F, B2, 1);
      if (B2 >= $80) then
      begin
        B2 := B2 - $80;
        BlockRead(F, B3, 1);
        if (B3 >= $80) then
        begin
          B3 := B3 - $80;
          BlockRead(F, B4, 1);
        end;
      end;
    end;
  end;

  //Assign DWORDs
  D0 := B0;
  D1 := B1;
  D2 := B2;
  D3 := B3;
  D4 := B4;
  //Shift Bytes
  D1 := D1 shl 6;
  D2 := D2 shl 13;
  D3 := D3 shl 20;
  D4 := D4 shl 27;

  //Create resulting DWORD
  INDEX := D0 + D1 + D2 + D3 + D4;

  //Check for sign
  if (IsNegative = TRUE) then INDEX := INDEX * (-1);

  Result := INDEX;
end;

procedure WriteIndex;
var
  B0,B1,B2,B3,B4 : Variant;
  D0,D1,D2,D3,D4 : DWORD;
  length         : integer;
  IsNegative     : Boolean;
  Value          : DWORD;
begin
  //check sign and remove it from original DWORD if necessary
  if ( Index < 0 ) then
  begin
    Value := Index * -1;
    IsNegative := TRUE;
  end
  else
  begin
    Value := Index;
    IsNegative := FALSE;
  end;

  //Create BYTEs from DWORD
  B0 := BYTE(Value);
  if ( B0 >= $80 ) then
    B0 := B0 - $80;
  if ( B0 >= $40 ) then
    B0 := B0 - $40;
  Value := Value shr 6;
  B1 := BYTE(Value);
  if ( B1 >= $80 ) then
    B1 := B1 - $80;
  Value := Value shr 7;
  B2 := BYTE(Value);
  if ( B2 >= $80 ) then
    B2 := B2 - $80;
  Value := Value shr 7;
  B3 := BYTE(Value);
  if ( B3 >= $80 ) then
    B3 := B3 - $80;
  Value := Value shr 7;
  B4 := BYTE(Value);
  if ( B4 >= $80 ) then
    B4 := B4- $80;

  //Assign sign-bit
  if ( IsNegative ) then
    B0 := B0 + $80;

  //Assign length-bits
  length := 1;
  if ( B1 + B2 + B3 + B4 > 0 ) then
  begin
    B0 := B0 + $40;
    length := length + 1;
    if ( B2 + B3 + B4 > 0 ) then
    begin
      B1 := B1 + $80;
      length := length + 1;
      if (  B3 + B4 > 0 ) then
      begin
        B2 := B2 + $80;
        length := length + 1;
        if ( B4 > 0 ) then
        begin
          length := length + 1;
          B3 := B3 + $80;
        end;
      end;
    end;
  end;

  D0 := B0;
  //Write INDEX to disk
  BlockWrite(F, D0, 1);
  if ( length > 1 ) then
  begin
    D1 := B1;
    BlockWrite(F, D1, 1);
    if ( length > 2 ) then
    begin
      D2 := B2;
      BlockWrite(F, D2, 1);
      if ( length > 3 ) then
      begin
        D3 := B3;
        BlockWrite(F, D3, 1);
        if ( length > 4 ) then
        begin
          D4 := B4;
          BlockWrite(F, D4, 1);
        end;
      end;
    end;
  end;
end;

function SizeOfINDEX;
var
  B0,B1,B2,B3,B4 : BYTE;
  length         : integer;
  Value          : DWORD;
begin
  //Create BYTEs
  if ( Index < 0 ) then
  begin
    Value := Index * -1;
  end
  else
  begin
    Value := Index;
  end;

  //Create BYTEs from DWORD
  B0 := BYTE(Value);
  if ( B0 >= $80 ) then
    B0 := B0 - $80;
  if ( B0 >= $40 ) then
    B0 := B0 - $40;
  Value := Value shr 6;
  B1 := BYTE(Value);
  if ( B1 >= $80 ) then
    B1 := B1 - $80;
  Value := Value shr 7;
  B2 := BYTE(Value);
  if ( B2 >= $80 ) then
    B2 := B2 - $80;
  Value := Value shr 7;
  B3 := BYTE(Value);
  if ( B3 >= $80 ) then
    B3 := B3 - $80;
  Value := Value shr 7;
  B4 := BYTE(Value);
  if ( B4 >= $80 ) then
    B4 := B4- $80;

  //length
  length := 1;
  if ( B1 > 0 ) then
  begin
    length := length + 1;
    if ( B2 > 0 ) then
    begin
      length := length + 1;
      if (  B3 > 0 ) then
      begin
        length := length + 1;
        if ( B4 > 0 ) then
        begin
          length := length + 1;
        end;
      end;
    end;
  end;

  Result := length;
end;

///////////////////////
//Read NAME-Format
function ReadNAME;
var
  zero : BYTE;
  i    : integer;
  NAME : string;
  temp : char;
begin
  i := 0;
  if (PkgVer < 64) then
  begin
    repeat
    i := i + 1;
    BlockRead(F, zero, 1);
    if (zero <> $00) then
    begin
      Seek(F, FilePos(F) - 1);
      BlockRead(F, temp, 1);
      Insert(temp, NAME, i);
    end;
    until (zero = $00);
  end
  else
  begin
    Seek(F, FilePos(F) + 1); //quit length byte
    repeat
    i := i + 1;
    BlockRead(F, zero, 1);
    if (zero <> $00) then
    begin
      Seek(F, FilePos(F) - 1);
      BlockRead(F, temp, 1);
      Insert(temp, NAME, i);
    end;
    until (zero = $00);
  end;

  Result := NAME;
end;

//supports only PkgVer 69
procedure WriteName;
var
  tempBYTE : BYTE;
  tempCHAR : char;
  i        : integer;
begin
  tempBYTE := length(Name) + 1;
  BlockWrite(F, tempBYTE, 1);
  for i:=0 to length(Name)-1 do
  begin
    tempCHAR := Name[i+1];
    BlockWrite(F, tempCHAR, 1);
  end;
  tempBYTE := $00;
  BlockWrite(F, tempBYTE, 1);
end;

////////////////////////////////////////////////////////////////////////////////
//Headertable-objects

///////////////////////
//TNameTable
constructor TNameTable.Create;
begin
  ObjectName  := TStringList.Create;
  ObjectFlags := TStringList.Create;
end;

procedure TNameTable.Free;
begin
 if Self <> nil then Destroy;
end;

destructor TNameTable.Destroy;
begin
  ObjectName.Free;
  ObjectFlags.Free;

  inherited Destroy;
end;

///////////////////////
//TImportTable
constructor TImportTable.Create;
begin
    ClassPackage   := TStringList.Create;
    ClassName      := TStringList.Create;
    ImportPackage  := TStringList.Create;
    ObjectName     := TStringList.Create;
end;

procedure TImportTable.Free;
begin
 if Self <> nil then Destroy;
end;

destructor TImportTable.Destroy;
begin
  ClassPackage.Free;
  ClassName.Free;
  ImportPackage.Free;
  ObjectName.Free;

  inherited Destroy;
end;

///////////////////////
//TExportTable
constructor TExportTable.Create;
begin
    ExportClass  := TStringList.Create;
    Super        := TStringList.Create;
    ExportPackage:= TStringList.Create;
    ObjectName   := TStringList.Create;
    ObjectFlags  := TStringList.Create;
    SerialSize   := TStringList.Create;
    SerialOffset := TStringList.Create;
end;

procedure TExportTable.Free;
begin
 if Self <> nil then Destroy;
end;

destructor TExportTable.Destroy;
begin
  ExportClass.Free;
  Super.Free;
  ExportPackage.Free;
  ObjectName.Free;
  ObjectFlags.Free;
  SerialSize.Free;
  SerialOffset.Free;

  inherited Destroy;
end;

//Evaluate Object-Reference
function GetObjectReference;
begin
  if (Value = 0) then Result := '';
  if (Value < 0) then Result := ImportTable.ObjectName.Strings[-Value-1];
  if (Value > 0) then Result := ExportTable.ObjectName.Strings[Value-1];
end;

////////////////////////////////////////////////////////////////////////////////
//Palette and color object

///////////////////////
//TUTColor
constructor TUTColor.Create;
begin
end;

procedure TUTColor.Free;
begin
 if Self <> nil then Destroy;
end;

destructor TUTColor.Destroy;
begin
  inherited Destroy;
end;

///////////////////////
//TPalette (UTPalette AND WindowsPalette)
constructor TPalette.Create;
var
  i         : integer;
begin
  for i:=0 to 255 do
    Color[i] := TUTColor.Create;
end;

procedure TPalette.Free;
begin
 if Self <> nil then Destroy;
end;

destructor TPalette.Destroy;
var
  i         : integer;
begin
  for i:=0 to 255 do
    Color[i].Free;

  inherited Destroy;
end;

procedure TPalette.Load;
var
  i       : integer;
  pal     : PLogPalette;
begin
  PaletteOffset := 0;
  PaletteName := Name;

  //Get offset of palette
  //palette and texture are in the same package
  for i:=0 to ExportTable.Count-1 do
    if (PaletteName = ExportTable.ObjectName[i]) then
      if (ExportTable.ExportClass.Strings[i] = 'Palette') then
        if (ExportTable.ExportPackage.Strings[i] = TexturePackage) then
          PaletteOffset := StrToInt(ExportTable.SerialOffset[i]);

  //palette and texture are in different packages
  if (PaletteOffset = 0) then
    for i:=0 to ExportTable.Count-1 do
      if (PaletteName = ExportTable.ObjectName[i]) then
        if (ExportTable.ExportClass.Strings[i] = 'Palette') then
          PaletteOffset := StrToInt(ExportTable.SerialOffset[i]);

  Seek(F, PaletteOffset);

  //Decode properties (there should be only a none-index)
  ReadINDEX(F);

  //Read the palette
  PaletteSize := ReadINDEX(F);
  for i:=0 to PaletteSize-1 do
  begin
    BlockRead(F, Color[i].r, 1);
    BlockRead(F, Color[i].g, 1);
    BlockRead(F, Color[i].b, 1);
    BlockRead(F, Color[i].a, 1);
  end;

  //Create palette
  pal := nil;
  try
    GetMem(pal, sizeof(TLogPalette) + sizeof(TPaletteEntry) * PaletteSize);
    pal.palVersion := $300;
    pal.palNumEntries := PaletteSize;
    for i := 0 to PaletteSize-1 do
    begin
      pal.palPalEntry[i].peRed := Color[i].r;
      pal.palPalEntry[i].peGreen := Color[i].g;
      pal.palPalEntry[i].peBlue := Color[i].b;
      pal.palPalEntry[i].peFlags := 0;
    end;
    WinPalette := CreatePalette(pal^);
  finally
    FreeMem(pal);
  end;
end;

////////////////////////////////////////////////////////////////////////////////
//TAddGroup

constructor TAddGroup.Create;
begin
  TextureName := TStringList.Create;
  FileName    := TStringList.Create;
  Offset      := TStringList.Create;
  HasMipMaps  := TStringList.Create;
  bMasked     := TStringList.Create;
end;

procedure TAddGroup.Free;
begin
 if Self <> nil then Destroy;
end;

destructor TAddGroup.Destroy;
begin
  TextureName.Free;
  FileName.Free;
  Offset.Free;
  HasMipMaps.Free;
  bMasked.Free;

  inherited Destroy;
end;

////////////////////////////////////////////////////////////////////////////////

end.
