{ Marco Cantu' (author of Master Delphi x and Delphi Developers Handbook has   }
{ released an excellent run-time object inspector tool as freeware.  It can be }
{ found in the Delphi Developers Handbook (highly recommended) and on his web  }
{ site at http://www.marcocantu.com.  If you have this installed, you can      }
{ enable the following define (remove the period) to use it in the demo.       }
{.$DEFINE DFS_RUN_TIME_DEBUGGER}

unit Unit1;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  {$IFDEF DFS_RUN_TIME_DEBUGGER}
  DdhObjDb,
  {$ENDIF}
  BrowseDr, StdCtrls, ExtCtrls, ComCtrls, CommCtrl, ShellAPI, ShlObj;

type
  TForm1 = class(TForm)
    BrowseDirectoryDlg: TdfsBrowseDirectoryDlg;
    Button2: TButton;
    Button1: TButton;
    ImageList1: TImageList;
    ListView1: TListView;
    Panel1: TPanel;
    Label1: TLabel;
    procedure Button2Click(Sender: TObject);
    procedure Button1Click(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure AddSelectionToImageList(Index: integer; const Path: string);
    procedure BrowseDirectoryDlgSelChanged(Sender: TObject; NewSel: string;
      NewSelPIDL: PItemIDList);
    procedure BrowseDirectoryDlgValidateFailed(Sender: TObject;
      Path: string; var Cancel: Boolean);
  private
    { Private declarations }
  public
    { Public declarations }
  end;

var
  Form1: TForm1;

implementation

{$R *.DFM}

uses
  FileCtrl;

procedure TForm1.Button2Click(Sender: TObject);
begin
  Close;
end;

procedure TForm1.Button1Click(Sender: TObject);
begin
  if BrowseDirectoryDlg.Execute then
  begin
    MessageDlg('You selected ' + BrowseDirectoryDlg.Selection + #13'The system'+
       ' image list index for this item is ' +
       IntToStr(BrowseDirectoryDlg.ImageIndex), mtInformation, [mbOk], 0);
    AddSelectionToImageList(BrowseDirectoryDlg.ImageIndex,
       BrowseDirectoryDlg.Selection);
  end;
end;

procedure TForm1.FormCreate(Sender: TObject);
begin
  {$IFDEF DFS_RUN_TIME_DEBUGGER}
  TDdhObjectDebugger.Create(Self);
  {$ENDIF}
  Panel1.Visible := FALSE;
  // Note that you can set the path you want to be initially selected in the
  // dialog by setting the Selection property before displayingthe dialog.  You
  // can also set this property at any time while the dialog is displayed and it
  // will reflect the change.
  BrowseDirectoryDlg.Selection := ExtractFilePath(Application.EXEName);
end;


// This procedure shows you how you can do something useful with the selected
// system image list index.  It retreives the image from the system image list
// and adds it our own image list, and then displays it in the listview.
procedure TForm1.AddSelectionToImageList(Index: integer; const Path: string);
var
  SFI: TSHFileInfo;
  SysIL: HImageList;
  SysImages: TImageList;
  Image: TBitmap;
  NewImgIdx: integer;
begin
  // Create the system image list.  We don't really "create" one.  We fool
  // SHGetFileInfo into giving us the system's image list and use it's handle.
  // You'd think it would be easier to get the system image list....
  SysImages := TImageList.Create(Self);
  Image := TBitmap.Create;
  try
    SysIL := SHGetFileInfo('', 0, SFI, SizeOf(SFI), SHGFI_SYSICONINDEX or
       SHGFI_SMALLICON);
    if SysIL <> 0 then
      with SysImages do
      begin
        Width := GetSystemMetrics(SM_CXSMICON);
        Height := GetSystemMetrics(SM_CYSMICON);
        Handle := SysIL;
        ShareImages := TRUE; // DON'T FREE THE SYSTEM IMAGE LIST! BAD IDEA (tm)!
      end;

    with ImageList1 do
    begin
      // Make sure our image list is the same as the systems.
      Width := GetSystemMetrics(SM_CXSMICON);
      Height := GetSystemMetrics(SM_CYSMICON);
      // Get the image from the system image list.
      SysImages.GetBitmap(Index, Image);
      // Add the given image to our list
      NewImgIdx := Add(Image, NIL);
      // Add something to the list view to display our new image.
      with ListView1.Items.Add do
      begin
        Caption := Path;
        ImageIndex := NewImgIdx;
      end;
    end;
  finally
    Image.Free;
    SysImages.Free;
  end;
end;


procedure TForm1.BrowseDirectoryDlgSelChanged(Sender: TObject;
  NewSel: string; NewSelPIDL: PItemIDList);
begin
  BrowseDirectoryDlg.EnableOKButton := (NewSel <> '');
end;

procedure TForm1.BrowseDirectoryDlgValidateFailed(Sender: TObject;
  Path: string; var Cancel: Boolean);
begin
  case MessageDlg(Path + ' does not exist, would you like to create it?',
     mtConfirmation, mbYesNoCancel, 0) of
    mrYes:
      begin
        Cancel := FALSE;
        try
          // This will create all the directories, but note that it will NOT
          // cause the treeview to be refreshed.  I don't know of any way to
          // get it to refresh the list.
          FileCtrl.ForceDirectories(Path);
          BrowseDirectoryDlg.Selection := Path;
        except
          ShowMessage('Failed to create direcotry ' + Path);
        end
      end;
    mrNo:
      Cancel := FALSE;
    mrCancel:
      Cancel := TRUE;
  end;
end;

end.
